!==================================================
! aLUA :   c LUA      2003
!==================================================
! :     LUA
! :
! 1)     - .
! 2)       ,  
!       - WRITE(*,*);
!
!==================================================
! VIA 25.10.12:  
!==================================================
MODULE aForLua
  use lua_interface
  USE iso_c_binding, only: C_CHAR, C_NULL_CHAR, C_INT, C_PTR, &
                           C_FUNPTR, C_DOUBLE, C_NULL_PTR

  IMPLICIT NONE
  character(len=4096),private:: EMSG

!--------------------------------------------------
! 1)   TALUA
!--------------------------------------------------
  Type :: TAForLUA
    type(c_ptr), private :: mLuaState
    integer, private :: NoFunArg      !   ,    
  contains
    procedure :: Create => TAForLUA_Create
    procedure :: Free => TAForLUA_Free
    procedure :: LoadScript => TAForLUA_LoadScript
    procedure :: DoScript => TAForLUA_DoScript
    procedure :: IsFunExists => TAForLUA_IsFunExists
!   ---      LUA---
    procedure :: TAForLUA_GetValReal
    procedure :: TAForLUA_GetValInt
    procedure :: TAForLUA_GetValChar
    procedure :: TAForLUA_GetValBool
    procedure :: TAForLUA_GetArrayReal
    GENERIC :: GetVal => TAForLUA_GetValReal,TAForLUA_GetValInt,TAForLUA_GetValChar, &
                         TAForLUA_GetValBool, TAForLUA_GetArrayReal
!   ---      LUA---
    procedure :: TAForLUA_SetValReal
    procedure :: TAForLUA_SetValInt
    procedure :: TAForLUA_SetValChar
    procedure :: TAForLUA_SetValBool
    procedure :: TAForLUA_SetArrayReal
    GENERIC :: SetVal => TAForLUA_SetValReal,TAForLUA_SetValInt,TAForLUA_SetValChar, &
                         TAForLUA_SetValBool, TAForLUA_SetArrayReal

!   ---     LUA---
    procedure :: OpenTable
    procedure :: CloseTable
    procedure :: TAForLUA_SetTableEntryReal
    procedure :: TAForLUA_SetTableEntryInt
    procedure :: TAForLUA_SetTableEntryChar
    procedure :: TAForLUA_SetTableEntryBool
    GENERIC :: SetTableEntry => TAForLUA_SetTableEntryReal,TAForLUA_SetTableEntryInt, &
                           TAForLUA_SetTableEntryChar, TAForLUA_SetTableEntryBool


!   ---     ---
    procedure :: SetFun => TAForLUA_SetFun
    procedure :: DoFun => TAForLUA_DoFun
    procedure :: TAForLUA_SetArgReal
    procedure :: TAForLUA_SetArgInt
    procedure :: TAForLUA_SetArgChar
    procedure :: TAForLUA_SetArgBool
    GENERIC :: SetArg => TAForLUA_SetArgReal,TAForLUA_SetArgInt,TAForLUA_SetArgChar, &
                            TAForLUA_SetArgBool

 ! VIA 03.02.13:     
    procedure :: OpenTableArg
    procedure :: CloseTableArg
    procedure :: TAForLUA_SetTableArgReal
    procedure :: TAForLUA_SetTableArgInt
    procedure :: TAForLUA_SetTableArgChar
    procedure :: TAForLUA_SetTableArgBool
    GENERIC :: SetTableArg => TAForLUA_SetTableArgReal,TAForLUA_SetTableArgInt, &
                  TAForLUA_SetTableArgChar, TAForLUA_SetTableArgBool

    procedure :: TAForLUA_GetFunResultReal
    procedure :: TAForLUA_GetFunResultInt
    procedure :: TAForLUA_GetFunResultChar
    procedure :: TAForLUA_GetFunResultBool
    GENERIC :: GetFunResult => TAForLUA_GetFunResultReal, &
                                  TAForLUA_GetFunResultInt, &
                                  TAForLUA_GetFunResultChar,TAForLUA_GetFunResultBool
!  ------   -----
   procedure :: DEBUG


  end Type
!====================================
  CONTAINS
!====================================

!------------------------------------
!   TAForLUA_Create
!------------------------------------
subroutine TAForLUA_Create(this)
  class (TAForLUA) :: this
!-----
! 1)    LUA
!
  this%mLuaState=luaL_newstate()
  this%NoFunArg = -1   !   
!
! 2)    LUA
!
  CALL luaL_openlibs(this%mLuaState)
end subroutine

!------------------------------------
!   TAForLUA_Free
!------------------------------------
subroutine TAForLUA_Free(this)
  class (TAForLUA) :: this
!-----
  call lua_close(this%mLuaState)
  this%mLuaState = c_null_ptr
  this%NoFunArg = -1
end subroutine

!------------------------------------
!   TAForLUA_LoadScript
!------------------------------------
logical function TAForLUA_LoadScript(this, fname) result(RC)
  class (TAForLUA) :: this
  character(len=*) :: fname
  integer(c_int) :: filesuccess
  character(len=3) :: c_mode
!-----
!          .
!         
!    .
!
  c_mode = 'bt' // c_null_char
  filesuccess = luaL_loadfilex(this%mLuaState, TRIM(fname)//C_NULL_CHAR, c_mode)
  if(filesuccess .EQ. 0) then
     rc = .TRUE.
  else
     rc = .FALSE.
     call flua_tolstring(this%mLuaState, -1, EMSG)
     write(*,*) 'aForLUA.LoadScript error: '//trim(EMSG)
  endif
  this%NoFunArg = -1
end function

!------------------------------------
!   TAForLUA_DoScript
!------------------------------------
logical function TAForLUA_DoScript(this) result(RC)
  class (TAForLUA) :: this
  integer(c_int) :: CallSuccess
!-----
!    ""  .
!
  CallSuccess = lua_pcallk(this%mLuaState,0,0,0,0_c_int, C_NULL_PTR)
  if (callsuccess.EQ.0) CALL lua_settop(this%mLuaState,-2)
  if(callsuccess .EQ. 0) then
     rc = .TRUE.
  else
     rc = .FALSE.
     call flua_tolstring(this%mLuaState, -1, EMSG)
     write(*,*) 'aForLUA.DoScript error: '//trim(EMSG)
  endif
end function



!------------------------------------
!   TAForLUA_GetValReal
!------------------------------------
logical function TAForLUA_GetValReal(this,VName,RVal) result(RC)
  class (TAForLUA) :: this
  real(4) RVal
  character(LEN=*) :: VName
  integer(kind=c_int) :: isnum
!-------------------------------
   RVAL = 0.0
   RC = .FALSE.
!  ---      LUA    VName
   call lua_getglobal(this%mLuaState,TRIM(VName)//C_NULL_CHAR)
!  ---       (   -1)
   if(lua_Type(this%mLuaState, -1) <= 0) return    !  
   if(lua_type(this%mLuaState,-1) .NE. LUA_TNUMBER ) return  !   
!  ---     (Index = -1)  
   RVAL = real(lua_tonumberx(this%mLuaState,-1,isnum),KIND=kind(RVAL))
!  ---        (-2)
!           
   call lua_settop(this%mLuaState,-2)
   RC = .TRUE.
end function

!------------------------------------
!   TAForLUA_GetValInt
!------------------------------------
logical function TAForLUA_GetValInt(this,VName,IVal) result(RC)
  class (TAForLUA) :: this
  integer IVal
  real(4) RVAL
  character(LEN=*) :: VName
!-------------------------------
   RC = TAForLUA_GetValReal(this,VName,RVal)
   IVAL = NINT(RVAL)
end function

!------------------------------------
!   TAForLUA_GetValChar
!------------------------------------
logical function TAForLUA_GetValChar(this,VName,CVal) result(RC)
  class (TAForLUA) :: this
  character(LEN=*) :: VName, CVal
!-------------------------------
   CVal = ' '
   RC = .FALSE.
   call lua_getglobal(this%mLuaState,TRIM(VName)//C_NULL_CHAR)
   if(lua_Type(this%mLuaState, -1) <= 0) return
   call flua_tolstring(this%mLuaState, -1, CVal)
   RC = .TRUE.
end function

!------------------------------------
!   TAForLUA_GetValBool
!------------------------------------
logical function TAForLUA_GetValBool(this,VName,LVal) result(RC)
  class (TAForLUA) :: this
  character(LEN=*) :: VName
  logical LVAL
!-------------------------------
   RC = .FALSE.
   call lua_getglobal(this%mLuaState,TRIM(VName)//C_NULL_CHAR)
   if(lua_Type(this%mLuaState, -1) <= 0) return
   if(lua_type(this%mLuaState,-1) .NE. LUA_TBOOLEAN ) return  !   LOG
   LVAL = lua_toboolean(this%mLuaState, -1).EQ.1
   RC = .TRUE.
end function



!------------------------------------
!   TAForLUA_SetValReal
!------------------------------------
subroutine TAForLUA_SetValReal(this,VName,RVal)
  class (TAForLUA) :: this
  real(4) RVal
  character(LEN=*) :: VName
  real(kind=c_double) luaVal
!-------------------------------
   LuaVal = RVAL
!  ---     
   call lua_pushnumber(this%mLuaState,LuaVal)
!  ---    VName   
   call lua_setglobal(this%mLuaState,TRIM(VName)//C_NULL_CHAR)
end subroutine

!------------------------------------
!   TAForLUA_SetValInt
!------------------------------------
subroutine TAForLUA_SetValInt(this,VName,IVal)
  class (TAForLUA) :: this
  integer  IVal
  character(LEN=*) :: VName
!-------------------------------
!  ---     
   call lua_pushinteger(this%mLuaState,IVal)
!  ---    VName   
   call lua_setglobal(this%mLuaState,TRIM(VName)//C_NULL_CHAR)
end subroutine

!------------------------------------
!   TAForLUA_SetValChar
!------------------------------------
subroutine TAForLUA_SetValChar(this,VName,CVal)
  class (TAForLUA) :: this
  character(LEN=*) :: VName,CVAL
!-------------------------------
!  ---     
   call flua_pushLstring(this%mLuaState,CVal)
!  ---    VName   
   call lua_setglobal(this%mLuaState,TRIM(VName)//C_NULL_CHAR)
end subroutine

!------------------------------------
!   TAForLUA_SetValBool
!------------------------------------
subroutine TAForLUA_SetValBool(this,VName,LVal)
  class (TAForLUA) :: this
  logical LVal
  character(LEN=*) :: VName
  integer(kind=c_int) :: iVal
!-------------------------------
   IVAL = 0
   if(LVAL) IVal = 1
!  ---     
   call lua_pushboolean(this%mLuaState,IVal)
!  ---    VName   
   call lua_setglobal(this%mLuaState,TRIM(VName)//C_NULL_CHAR)
end subroutine


!------------------------------------
!   TAForLUA_GetArrayReal
!------------------------------------
logical function TAForLUA_GetArrayReal(this,VName,RVal,LenARR) result(RC)
  class (TAForLUA) :: this
  real(4) RR, RVal(*)
  character(LEN=*) :: VName
  integer(kind=c_int) :: isnum
  integer LenARR,iPOS
!-------------------------------
   RVAL(1:LenARR) = 0.0
   RC = .FALSE.
!  ---      LUA    VName
   call lua_getglobal(this%mLuaState,TRIM(VName)//C_NULL_CHAR)
!  ---       (   -1)
   if(lua_Type(this%mLuaState, -1) <= 0) return    !  
   if(lua_Type(this%mLuaState,-1) .NE. LUA_TTABLE ) return  !   
!  ---
!       ""     -   
!  ---
!   ,  ,   NUMBER
   iPOS = 0
   CALL lua_pushnil(this%mLuaState)   !   _next
   DO WHILE(lua_next(this%mLuaState, -2).NE.0)
!   ---    'key' (at index -2) and 'value' (at index -1)
    iPOS = IPOS + 1
    if(lua_type(this%mLuaState,-1) .EQ. LUA_TNUMBER) then  ! value - 
       RR = real(lua_tonumberx(this%mLuaState,-1,isnum),KIND=kind(RVAL))
       if(iPOS.LE.LenARR) RVAL(iPOS) = RR
    endif
    CALL flua_pop(this%mLuaState, 1)     !  Value, key   
   END DO
   if(iPOS.LE.LenARR) LenARR = iPOS
   RC = .TRUE.
end function


!------------------------------------
!   TAForLUA_SetArrayReal
!------------------------------------
subroutine TAForLUA_SetArrayReal(this,VName,RVal,LenARR)
  class (TAForLUA) :: this
  real(4) RVal(*)
  character(LEN=*) :: VName
  integer LenARR,i
!-------------------------------
!  ---    
  call lua_createtable(this%mLuaState,0,0)
!
!write(*,*) 'Type(CreateTable)=',lua_type(this%mLuaState,-1)
!
!  ---  
  do i=1,LenARR
     call flua_setArrayField(this%mLuaState, i,RVal(i))
  enddo
!
!
!write(*,*) 'Type(SetGlobal)=',lua_type(this%mLuaState,-1)
!
  call lua_setglobal(this%mLuaState, TRIM(VName)//C_NULL_CHAR)
!
end subroutine


!=====================================
!     
!=====================================

!------------------------------------
!   TAForLUA_OpenFun
!------------------------------------
logical function TAForLUA_SetFun(this,FunName) result(RC)
  class (TAForLUA) :: this
  character(LEN=*) :: FunName
!-------------------------------
   RC = .FALSE.
!  ---      LUA    FunName
   call lua_getglobal(this%mLuaState,TRIM(FunName)//C_NULL_CHAR)
!  ---       (   -1)
   if(lua_Type(this%mLuaState, -1) <= 0) return    !  
   if(lua_Type(this%mLuaState,-1) .NE. LUA_TFUNCTION ) return  !   
   RC = .TRUE.
   this%NoFunArg = 0
end function

!------------------------------------
!   TAForLUA_DoFun
!------------------------------------
logical function TAForLUA_DoFun(this) result(RC)
  class (TAForLUA) :: this
  integer(c_int) :: CallSuccess
  integer(kind=c_int) :: c_nargs
  integer(kind=c_int) :: c_nresults
!-------
!   (    )
!-------
  RC = .FALSE.
  if(this%NoFunArg.EQ.-1) return !     
  c_nargs =  this%NoFunArg
  c_nresults = 1 !       .
! -----
  CallSuccess = lua_pcallk(this%mLuaState,c_nargs,c_nresults,0,0_c_int, C_NULL_PTR)
! -----
  if(callsuccess .EQ. 0) then
     rc = .TRUE.
  else
     rc = .FALSE.
     call flua_tolstring(this%mLuaState, -1, EMSG)
     write(*,*) 'aForLUA.DoScript error: '//trim(EMSG)
  endif
end function


!------------------------------------
!   TAForLUA_SetArgReal
!------------------------------------
subroutine TAForLUA_SetArgReal(this,RVal)
  class (TAForLUA) :: this
  real(4) RVal
  real(kind=c_double) luaVal
!-------------------------------
   LuaVal = RVAL
   call lua_pushnumber(this%mLuaState,LuaVal)
   this%NoFunArg = this%NoFunArg + 1
end subroutine

!------------------------------------
!   TAForLUA_SetArgInt
!------------------------------------
subroutine TAForLUA_SetArgInt(this,IVal)
  class (TAForLUA) :: this
  integer IVal
  integer(kind=c_int) luaVal
!-------------------------------
   LuaVal = IVAL
   call lua_pushinteger(this%mLuaState,LuaVal)
   this%NoFunArg = this%NoFunArg + 1
end subroutine

!------------------------------------
!   TAForLUA_SetArgChar
!------------------------------------
subroutine TAForLUA_SetArgChar(this,CVal)
  class (TAForLUA) :: this
  character(LEN=*) :: CVAL
!-------------------------------
   call flua_pushLstring(this%mLuaState,CVal)
   this%NoFunArg = this%NoFunArg + 1
end subroutine

!------------------------------------
!   TAForLUA_SetArgBool
!------------------------------------
subroutine TAForLUA_SetArgBool(this,LVal)
  class (TAForLUA) :: this
  logical LVal
  integer(kind=c_int) :: iVal
!-------------------------------
   IVAL = 0
   if(LVAL) IVal = 1
   call lua_pushboolean(this%mLuaState,IVal)
   this%NoFunArg = this%NoFunArg + 1
end subroutine


!------------------------------------
!   TAForLUA_GetFunResultReal
!------------------------------------
logical function TAForLUA_GetFunResultReal(this,RVal) result(RC)
  class (TAForLUA) :: this
  real(4) RVal
  integer(kind=c_int) :: isnum
!-------------------------------
   RVAL = 0.0
   RC = .FALSE.
   if(this%NoFunArg.EQ.-1) return !     
   if(lua_Type(this%mLuaState, -1) <= 0) return    !  
   if(lua_isnumber(this%mLuaState,-1) .EQ. 0 ) return  !   
   RVAL = real(lua_tonumberx(this%mLuaState,-1,isnum),KIND=kind(RVAL))
   call lua_settop(this%mLuaState,-2)
   RC = .TRUE.
   this%NoFunArg = -1   !  
end function

!------------------------------------
!   TAForLUA_GetFunResultInt
!------------------------------------
logical function TAForLUA_GetFunResultInt(this,IVal) result(RC)
  class (TAForLUA) :: this
  real(4) RVal
  integer IVal
!-------------------------------
   IVAL = 0
   RC = .FALSE.
   if(this%NoFunArg.EQ.-1) return !     
   RC = TAForLUA_GetFunResultReal(this,RVal)
   IVAL = NINT(RVal)
   this%NoFunArg = -1   !  
end function

!------------------------------------
!   TAForLUA_FunResultChar
!------------------------------------
logical function TAForLUA_GetFunResultChar(this,CVal) result(RC)
  class (TAForLUA) :: this
  character(LEN=*) :: CVal
!-------------------------------
   CVal = ' '
   RC = .FALSE.
   if(this%NoFunArg.EQ.-1) return !     
   if(lua_Type(this%mLuaState, -1) <= 0) return
   call flua_tolstring(this%mLuaState, -1, CVal)
   RC = .TRUE.
   this%NoFunArg = -1   !  
end function

!------------------------------------
!   TAForLUA_FunResultBool
!------------------------------------
logical function TAForLUA_GetFunResultBool(this,LVal) result(RC)
  class (TAForLUA) :: this
  logical LVAL
!-------------------------------
   RC = .FALSE.
   if(this%NoFunArg.EQ.-1) return !     
   if(lua_Type(this%mLuaState, -1) <= 0) return
   if(lua_type(this%mLuaState,-1) .NE. LUA_TBOOLEAN ) return  !   LOG
   LVAL = lua_toboolean(this%mLuaState, -1).EQ.1
   RC = .TRUE.
   this%NoFunArg = -1   !  
end function

logical function TAForLUA_IsFunExists(this,FunName)  result(RC)
  class (TAForLUA) :: this
  character(LEN=*) :: FunName
! -----
  RC = .false.
  call lua_getglobal(this%mLuaState,TRIM(FunName)//C_NULL_CHAR)
  if(lua_Type(this%mLuaState, -1) <= 0) return
  if(lua_Type(this%mLuaState,-1) .NE. LUA_TFUNCTION ) return
   RC = .TRUE.
end function

!------------------------------------
!   OpenTable
!------------------------------------
subroutine OpenTable(this)
  class (TAForLUA) :: this
!-------------------------------
!  ---    
!
  call lua_createtable(this%mLuaState,0,0)
end subroutine

!------------------------------------
!   CloseTable
!------------------------------------
subroutine CloseTable(this,TName)
  class (TAForLUA) :: this
  character(LEN=*) :: TName
!-------------------------------
!  ---   
!
  call lua_setglobal(this%mLuaState, TRIM(TName)//C_NULL_CHAR)
end subroutine


!------------------------------------
!   TAForLUA_SetTableEntryReal
!------------------------------------
subroutine TAForLUA_SetTableEntryReal(this,EName,RVal)
  class (TAForLUA) :: this
  character(LEN=*) :: EName
  real(4) RVal
  real(kind=c_double) luaVal
!-------------------------------
   LuaVal = RVAL
   call  lua_pushstring(this%mLuaState,trim(EName)//C_NULL_CHAR)
   call  lua_pushnumber(this%mLuaState, LuaVal)
   call  lua_settable(this%mLuaState, -3);
end subroutine

!------------------------------------
!   TAForLUA_SetTableEntryInt
!------------------------------------
subroutine TAForLUA_SetTableEntryInt(this,EName,IVal)
  class (TAForLUA) :: this
  character(LEN=*) :: EName
  integer IVal
!-------------------------------
   call  flua_pushLstring(this%mLuaState,EName)
   call  lua_pushinteger(this%mLuaState,IVal)
   call  lua_settable(this%mLuaState, -3);
end subroutine

!------------------------------------
!   TAForLUA_SetTableEntryBool
!------------------------------------
subroutine TAForLUA_SetTableEntryBool(this,EName,LVal)
  class (TAForLUA) :: this
  character(LEN=*) :: EName
  logical LVal
  integer(kind=c_int) :: iVal
!-------------------------------
   IVAL = 0
   if(LVAL) IVal = 1
   call  flua_pushLstring(this%mLuaState,EName)
   call  lua_pushboolean(this%mLuaState,IVal)
   call  lua_settable(this%mLuaState, -3);
end subroutine

!------------------------------------
!   TAForLUA_SetTableEntryChar
!------------------------------------
subroutine TAForLUA_SetTableEntryChar(this,EName,CVal)
  class (TAForLUA) :: this
  character(LEN=*) :: EName,CVAL
!-------------------------------
   call  flua_pushLstring(this%mLuaState,EName)
   call  flua_pushLstring(this%mLuaState,CVal)
   call  lua_settable(this%mLuaState, -3);
end subroutine

!------------------------------------
!   OpenTableArg
!------------------------------------
subroutine OpenTableArg(this)
  class (TAForLUA) :: this
!-------------------------------
!  ---    
!
  call lua_createtable(this%mLuaState,0,0)
end subroutine

!------------------------------------
!   CloseTableArg
!------------------------------------
subroutine CloseTableArg(this)
  class (TAForLUA) :: this
!-------------------------------
!  ---    
!
  this%NoFunArg = this%NoFunArg + 1
end subroutine

!------------------------------------
!   TAForLUA_SetTableArgReal
!------------------------------------
subroutine TAForLUA_SetTableArgReal(this,EName,RVal)
  class (TAForLUA) :: this
  character(*) EName
  real(4) RVal
!-------------------------------
   call TAForLUA_SetTableEntryReal(this,EName,RVal)
end subroutine

!------------------------------------
!   TAForLUA_SetTableArgInt
!------------------------------------
subroutine TAForLUA_SetTableArgInt(this,EName,IVal)
  class (TAForLUA) :: this
  character(*) EName
  integer IVal
!-------------------------------
   call TAForLUA_SetTableEntryInt(this,EName,IVal)
end subroutine

!------------------------------------
!   TAForLUA_SetTableArgChar
!------------------------------------
subroutine TAForLUA_SetTableArgChar(this,EName,CVal)
  class (TAForLUA) :: this
  character(*) EName, CVal
!-------------------------------
   call TAForLUA_SetTableEntryChar(this,EName,CVal)
end subroutine

!------------------------------------
!   TAForLUA_SetTableArgBool
!------------------------------------
subroutine TAForLUA_SetTableArgBool(this,EName,LVal)
  class (TAForLUA) :: this
  character(*) EName
  logical LVal
!-------------------------------
   call TAForLUA_SetTableEntryBool(this,EName,LVal)
end subroutine



!========================================
!    
!========================================
subroutine DEBUG(this)
  class (TAForLUA) :: this
!-----
!// TODO ():    TO DO
!  call lua_createtable(this%mLuaState,0,0)
!  write(*,*) 'Type(CreateTable)=',lua_type(this%mLuaState,-1)
!   call  lua_pushinteger(this%mLuaState, 1);
!!   call  lua_pushstring(this%mLuaState, 'a'//char(0));
!  write(*,*) 'Type(pushinteger)=',lua_type(this%mLuaState,-1)
!   call  lua_pushnumber(this%mLuaState, 1.0d+01);
!  write(*,*) 'Type(pushnumber)=',lua_type(this%mLuaState,-1)
!   call  lua_settable(this%mLuaState, -3);
!  write(*,*) 'Type(settable)=',lua_type(this%mLuaState,-1)
!
!
!  call lua_setglobal(this%mLuaState, 'pipi'//C_NULL_CHAR)
end subroutine



END MODULE
